import datetime

__all__ = [
    "get_current_datetime_extra_data",
]


weekdays = [
    ["星期一", "周一", "礼拜一"],
    ["星期二", "周二", "礼拜二"],
    ["星期三", "周三", "礼拜三"],
    ["星期四", "周四", "礼拜四"],
    ["星期五", "周五", "礼拜五"],
    ["星期六", "周六", "礼拜六"],
    ["星期日", "星期天", "周日", "礼拜天", "礼拜日"],
]


def get_weekday_info(nowtime: datetime.datetime, prefix: str = "今天是"):
    return "，".join([prefix + x for x in weekdays[nowtime.weekday()]]) + "。"


def get_day_info(nowtime: datetime.datetime, prefix: str = "今天是"):
    return prefix + nowtime.strftime("%Y-%m-%d") + "。"


def get_nowtime(nowtime: datetime.datetime, prefix: str = "现在时间是"):
    return prefix + nowtime.strftime("%Y-%m-%d %H:%M:%S") + "。"


def get_week_range_info(nowtime: datetime.datetime, prefix: str = "本周时间范围是"):
    today_weekday = nowtime.weekday()
    monday = nowtime - datetime.timedelta(days=today_weekday)
    sunday = monday + datetime.timedelta(days=6)
    return (
        prefix
        + "从"
        + monday.strftime("%Y-%m-%d 00:00:00")
        + "到"
        + sunday.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_this_month_range_info(
    nowtime: datetime.datetime, prefix: str = "本月时间范围是"
):
    next_month_year = nowtime.year
    next_month_month = nowtime.month + 1
    if next_month_month > 12:
        next_month_month = 1
        next_month_year += 1
    day1 = datetime.datetime(nowtime.year, nowtime.month, 1, 0, 0, 0)
    day31 = datetime.datetime(
        next_month_year, next_month_month, 1, 0, 0, 0
    ) - datetime.timedelta(days=1)
    return (
        prefix
        + "从"
        + day1.strftime("%Y-%m-%d 00:00:00")
        + "到"
        + day31.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_last_month_range_info(
    nowtime: datetime.datetime, prefix: str = "上月时间范围是"
):
    day31 = datetime.datetime(nowtime.year, nowtime.month, 1) - datetime.timedelta(
        days=1
    )
    day1 = datetime.datetime(day31.year, day31.month, 1, 0, 0, 0)
    return (
        prefix
        + "从"
        + day1.strftime("%Y-%m-%d 00:00:00")
        + "到"
        + day31.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_next_month_range_info(
    nowtime: datetime.datetime, prefix: str = "下个月时间范围是"
):
    next_month_year = nowtime.year
    next_month_month = nowtime.month + 1
    if next_month_month > 12:
        next_month_month = 1
        next_month_year += 1
    day1 = datetime.datetime(next_month_year, next_month_month, 1, 0, 0, 0)

    next_month_year = day1.year
    next_month_month = day1.month + 1
    if next_month_month > 12:
        next_month_month = 1
        next_month_year += 1
    day31 = datetime.datetime(
        next_month_year, next_month_month, 1
    ) - datetime.timedelta(days=1)
    return (
        prefix
        + "从"
        + day1.strftime("%Y-%m-%d 00:00:00")
        + "到"
        + day31.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_year_info(year: int, prefix: str = "今年是"):
    return f"{prefix}{year}年。"


def get_lastn_days_info(
    nowtime: datetime.datetime,
    n: int,
    prefix: str = "最近7天的范围是",
):
    start = nowtime - datetime.timedelta(days=n - 1)
    return (
        prefix
        + "从"
        + start.strftime("%Y-%m-%d 00:00:00")
        + "至"
        + nowtime.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_day_time_range(nowtime: datetime.datetime, prefix="今天的时间范围是"):
    return (
        prefix
        + nowtime.strftime("%Y-%m-%d 00:00:00")
        + "至"
        + nowtime.strftime("%Y-%m-%d 23:59:59")
    )


quarter_start_months = {
    1: 1,
    2: 1,
    3: 1,
    4: 3,
    5: 3,
    6: 3,
    7: 6,
    8: 6,
    9: 6,
    10: 9,
    11: 9,
    12: 9,
}


def get_this_quarter_range_info(
    nowtime: datetime.datetime, prefix="这个季度的时间范围是"
):
    start_month = quarter_start_months[nowtime.month]
    start_day = datetime.datetime(nowtime.year, start_month, 1)
    end_day = datetime.datetime(
        start_day.year, start_day.month + 3, 1
    ) - datetime.timedelta(days=1)
    return (
        prefix
        + start_day.strftime("%Y-%m-%d 00:00:00")
        + "至"
        + end_day.strftime("%Y-%m-%d 23:59:59")
    )


def get_last_quarter_range_info(
    nowtime: datetime.datetime, prefix="上个季度的时间范围是"
):
    if nowtime.month > 3:
        start_year = nowtime.year
        start_month = quarter_start_months[nowtime.month]
    else:
        start_year = nowtime.year - 1
        start_month = 9

    if start_month >= 9:
        end_year_nextday = start_year + 1
        end_month_nextday = 4
    else:
        end_year_nextday = start_year
        end_month_nextday = start_month + 3

    start_day = datetime.datetime(start_year, start_month, 1)
    end_day = datetime.datetime(
        end_year_nextday, end_month_nextday, 1
    ) - datetime.timedelta(days=1)
    return (
        prefix
        + start_day.strftime("%Y-%m-%d 00:00:00")
        + "至"
        + end_day.strftime("%Y-%m-%d 23:59:59")
    )


def get_pastn_days_info(nowtime: datetime.datetime, n=7, prefix="过去一周的时间范围是"):
    end_day = nowtime - datetime.timedelta(days=1)
    start_day = end_day - datetime.timedelta(days=n - 1)
    return (
        prefix
        + start_day.strftime("%Y-%m-%d 00:00:00")
        + "至"
        + end_day.strftime("%Y-%m-%d 23:59:59")
        + "。"
    )


def get_current_datetime_extra_data(nowtime=None):
    nowtime = nowtime or datetime.datetime.now()
    return [
        get_nowtime(nowtime=nowtime),
        get_day_info(nowtime=nowtime, prefix="今天是"),
        get_day_info(nowtime=nowtime + datetime.timedelta(days=1), prefix="明天是"),
        get_day_info(nowtime=nowtime - datetime.timedelta(days=1), prefix="昨天是"),
        get_weekday_info(nowtime=nowtime, prefix="今天是"),
        get_weekday_info(nowtime=nowtime + datetime.timedelta(days=1), prefix="明天是"),
        get_weekday_info(nowtime=nowtime - datetime.timedelta(days=1), prefix="昨天是"),
        get_year_info(year=nowtime.year, prefix="今年是"),
        get_year_info(year=nowtime.year - 1, prefix="去年是"),
        get_year_info(year=nowtime.year + 1, prefix="明年是"),
        get_pastn_days_info(nowtime=nowtime, n=7, prefix="过去一周的时间范围是"),
        get_pastn_days_info(nowtime=nowtime, n=1, prefix="过去一天的时间范围是"),
        get_pastn_days_info(nowtime=nowtime, n=3, prefix="过去三天的时间范围是"),
        get_pastn_days_info(nowtime=nowtime, n=7, prefix="过去七天的时间范围是"),
        get_lastn_days_info(nowtime=nowtime, n=7, prefix="最近一周的时间范围是"),
        get_lastn_days_info(nowtime=nowtime, n=7, prefix="最近7天的时间范围是"),
        get_lastn_days_info(nowtime=nowtime, n=14, prefix="最近14天的时间范围是"),
        get_lastn_days_info(nowtime=nowtime, n=21, prefix="最近21天的时间范围是"),
        get_lastn_days_info(nowtime=nowtime, n=30, prefix="最近30天的时间范围是"),
        get_day_time_range(nowtime=nowtime, prefix="今天的时间范围是"),
        get_day_time_range(
            nowtime=nowtime - datetime.timedelta(days=1), prefix="昨天的时间范围是"
        ),
        get_day_time_range(
            nowtime=nowtime + datetime.timedelta(days=1), prefix="明天的时间范围是"
        ),
        get_week_range_info(nowtime=nowtime, prefix="本周的时间范围是"),
        get_week_range_info(nowtime=nowtime, prefix="当前周的时间范围是"),
        get_week_range_info(nowtime=nowtime, prefix="这周的时间范围是"),
        get_week_range_info(nowtime=nowtime, prefix="这一周的时间范围是"),
        get_week_range_info(
            nowtime=nowtime - datetime.timedelta(days=7), prefix="上周的时间范围是"
        ),
        get_week_range_info(
            nowtime=nowtime - datetime.timedelta(days=7), prefix="上一周的时间范围是"
        ),
        get_week_range_info(
            nowtime=nowtime + datetime.timedelta(days=7), prefix="下周的时间范围是"
        ),
        get_week_range_info(
            nowtime=nowtime + datetime.timedelta(days=7), prefix="下一周的时间范围是"
        ),
        get_this_month_range_info(nowtime=nowtime, prefix="本月的时间范围是"),
        get_this_month_range_info(nowtime=nowtime, prefix="这月的时间范围是"),
        get_this_month_range_info(nowtime=nowtime, prefix="这个月的时间范围是"),
        get_last_month_range_info(nowtime=nowtime, prefix="上月的时间范围是"),
        get_last_month_range_info(nowtime=nowtime, prefix="上个月的时间范围是"),
        get_next_month_range_info(nowtime=nowtime, prefix="下月的时间范围是"),
        get_next_month_range_info(nowtime=nowtime, prefix="下个月的时间范围是"),
        get_this_quarter_range_info(nowtime=nowtime, prefix="这个季度的时间范围是"),
        get_this_quarter_range_info(nowtime=nowtime, prefix="本季度的时间范围是"),
        get_this_quarter_range_info(nowtime=nowtime, prefix="当前季度的时间范围是"),
        get_last_quarter_range_info(nowtime=nowtime, prefix="上个季度的时间范围是"),
        "重要：表示年份必须使用4位数字。禁止使用2位数字表示年份。",
        "重要：每天的时间范围总是从这一天的00:00:00到这一天的23:59:59。",
        "重要：最近(n)天，是指过去(n-1)天至今天。",
        "重要：过去(n)天，是指过去(n)天至昨天。",
        "重要：总是把星期一当成是一周的开始，把星期天当成是一周的结束。",
    ]



__all__ = [
    "ParseJsonResponseError",
    "ChatError",
    "PromptOrTemplateRequired",
    "GetTextEmbeddingsError",
    "GetRerankScoresError",
    "NoneValueError",
    "VectorStoreNoneValueError",
    "EmbeddingsNoneValueError",
    "RerankNoneValueError",
]


class ParseJsonResponseError(RuntimeError):
    """LLM输出不能解析为json数据。"""

    def __init__(self, *args, **kwargs):
        args = args or ["LLM输出不能解析为json数据。"]
        super().__init__(*args, **kwargs)


class ChatError(RuntimeError):
    """大模型对话失败。"""

    def __init__(self, *args, **kwargs):
        args = args or ["大模型对话失败。"]
        super().__init__(*args, **kwargs)


class PromptOrTemplateRequired(RuntimeError):
    """prompt或template不能同时为空。"""

    def __init__(self, *args, **kwargs):
        args = args or ["prompt或template不能同时为空。"]
        super().__init__(*args, **kwargs)


class GetTextEmbeddingsError(RuntimeError):
    """获取文本向量失败。"""

    def __init__(self, *args, **kwargs):
        args = args or ["获取文本向量失败。"]
        super().__init__(*args, **kwargs)


class GetRerankScoresError(RuntimeError):
    """获取rerank得分失败。"""

    def __init__(self, *args, **kwargs):
        args = args or ["获取rerank得分失败。"]
        super().__init__(*args, **kwargs)


class NoneValueError(RuntimeError):
    """不允许None值。"""

    def __init__(self, *args, **kwargs):
        args = args or ["不允许None值。"]
        super().__init__(*args, **kwargs)


class VectorStoreNoneValueError(NoneValueError):
    """向量数据库不允许None值进行查询和插入。"""

    def __init__(self, *args, **kwargs):
        args = args or ["向量数据库不允许None值进行查询和插入。"]
        super().__init__(*args, **kwargs)


class EmbeddingsNoneValueError(NoneValueError):
    """None值无法进行向量化处理。"""

    def __init__(self, *args, **kwargs):
        args = args or ["None值无法进行向量化处理。"]
        super().__init__(*args, **kwargs)


class RerankNoneValueError(NoneValueError):
    """None值无法参与重排。"""

    def __init__(self, *args, **kwargs):
        args = args or ["None值无法参与重排。"]
        super().__init__(*args, **kwargs)



from typing import Optional
from zenutils import importutils
from .base import get_template_engine
from .base import set_llmhelper_default_config

__all__ = [
    "get_template_prompt_by_django_template_engine",
    "get_template_prompt_by_jinjia2",
    "get_template_prompt",
]


def get_template_prompt_by_django_template_engine(
    template: Optional[str] = None,
    template_root: str = None,
    prompt: str = None,
    **context,
):
    """使用django模板引擎生成最终提示词。"""
    from django.template.loader import render_to_string

    return render_to_string(
        template,
        context={
            "prompt": prompt,
            **context,
        },
    )


def get_template_prompt_by_jinjia2(
    template: Optional[str] = None,
    template_root: str = None,
    prompt: str = None,
    **context,
):
    """使用jinja2模板引擎生成最终提示词。"""
    from jinja2 import Environment
    from jinja2 import FileSystemLoader

    template_root = template_root or "templates/"
    environment = Environment(loader=FileSystemLoader(template_root))
    tempalte = environment.get_template(template)
    return tempalte.render(prompt=prompt, **context)


def get_template_prompt(
    template: Optional[str] = None,
    prompt: str = None,
    template_root: Optional[str] = None,
    template_engine=None,
    **context,
):
    """根据提示词模板、用户问题和其它参数，生成最终的提示词。"""
    if template_engine:
        if callable(template_engine):
            return template_engine(
                template=template,
                template_root=template_root,
                prompt=prompt,
                **context,
            )
        else:
            template_engine = importutils.import_from_string(template_engine)
            return template_engine(
                template=template,
                template_root=template_root,
                prompt=prompt,
                **context,
            )
    else:
        template_engine = get_template_engine()
        if not template_engine:
            return get_template_prompt_by_django_template_engine(
                template=template,
                template_root=template_root,
                prompt=prompt,
                **context,
            )
        elif callable(template_engine):
            return template_engine(
                template=template,
                template_root=template_root,
                prompt=prompt,
                **context,
            )
        else:
            template_engine = importutils.import_from_string(template_engine)
            return template_engine(
                template=template,
                template_root=template_root,
                prompt=prompt,
                **context,
            )


set_llmhelper_default_config(
    template_engine=get_template_prompt_by_django_template_engine,
)



import logging
from typing import Dict
from typing import List

import yaml

from django.db import models
from django.db import transaction
from llmhelper.vectorstores import get_cached_vectorstore

__all__ = [
    "WithVectorStoreIndex",
]
_logger = logging.getLogger(__name__)


class WithVectorStoreIndex(models.Model):
    _cached_vectorstores = {}
    enable_vectorstore_index = True

    vectorstore_updated = models.BooleanField(
        null=True,
        blank=True,
        verbose_name="已更新向量数据库",
    )
    vectorstore_uids_data = models.TextField(
        null=True,
        blank=True,
        verbose_name="向量数据库记录编号",
        help_text="添加至向量数据库后返回的编号。用于后续向量数据库的数据维护。",
    )

    class Meta:
        abstract = True

    def get_enable_vectorstore_index_flag(self) -> bool:
        """用于判断该记录是否需要被索引。

        返回True表示需要被索引。
        返回False表示不需要被索引。
        """
        return True

    def get_vectorstore_index_names(self) -> List[str]:
        """返回该记录需要添加到哪个索引。"""
        raise NotImplementedError()

    def get_vectorstore_index_contents(self) -> List[str]:
        raise NotImplementedError()

    def get_vectorstore_index_metas(self) -> List[Dict[str, str]]:
        return []

    def get_vectorstore_uids(self):
        if not self.vectorstore_uids_data:
            return []
        else:
            return yaml.safe_load(self.vectorstore_uids_data)

    def set_vectorstore_uids(self, value):
        if not value:
            self.vectorstore_uids_data = None
        else:
            self.vectorstore_uids_data = yaml.safe_dump(value)

    vectorstore_uids = property(get_vectorstore_uids, set_vectorstore_uids)

    def save(self, *args, **kwargs):
        self.vectorstore_updated = False
        result = super().save(*args, **kwargs)
        transaction.on_commit(self.update_vectorstore_index)
        return result

    def update_vectorstore_index(self):
        try:
            if self.enable_vectorstore_index:
                if self.get_enable_vectorstore_index_flag():
                    self.upsert_index(save=False)
                else:
                    self.delete_index(save=False)
            else:
                self.delete_index(save=False)
            self.vectorstore_updated = True
        except Exception as error:
            self.vectorstore_updated = False
            _logger.error(
                "更新向量数据库失败：model=%s.%s, id=%s, error=%s",
                self._meta.app_label,
                self._meta.model_name,
                self.id,
                error,
            )
        self.__class__.objects.bulk_update(
            [self],
            fields=[
                "vectorstore_updated",
                "vectorstore_uids_data",
            ],
        )

    def delete_index(self, save=False):
        deleted = 0
        if self.vectorstore_uids:
            vs = get_cached_vectorstore()
            deleted += vs.delete_many(self.vectorstore_uids)
            self.vectorstore_uids = None
            if save:
                self.save()
        return deleted

    def upsert_index(self, save=False):
        vs_cotnents = self.get_vectorstore_index_contents()
        vs_metas = self.get_vectorstore_index_metas()
        # 先删除
        self.delete_index(save=False)
        # 后添加
        uids = []
        for index_name in self.get_vectorstore_index_names():
            vs = get_cached_vectorstore(index_name=index_name)
            uids += vs.insert_many(contents=vs_cotnents, metas=vs_metas)
        # 更新数据库记录
        self.vectorstore_uids = uids
        if save:
            self.save()



import re
import os
import json
import logging

from typing import List
from typing import Dict
from typing import Union
from typing import Optional
from typing import Tuple

import yaml
from openai import OpenAI

from .exceptions import ChatError
from .exceptions import PromptOrTemplateRequired
from .base import get_default_llm
from .base import get_default_chat_model
from .base import get_llm_chat_log_name
from .base import get_llm_chat_log_level
from .template import get_template_prompt
from .template import get_template_prompt_by_jinjia2

__all__ = [
    "get_messages",
    "parse_json_response",
    "chat",
    "jsonchat",
    "streaming_chat",
    "chat_with_extra_data",
    "jsonchat_with_extra_data",
    "streaming_chat_with_extra_data",
]
_logger = logging.getLogger(__name__)
_llm_chat_logger = logging.getLogger(get_llm_chat_log_name())
_llm_chat_log_level = get_llm_chat_log_level()


def get_messages(
    prompt: Union[str, List[Dict[str, str]]],
    histories: Optional[List[Tuple[str, str]]] = None,
    system_prompt: str = "You are helpful assistance.",
):
    """将文字版的prompt转化为messages数组。

    @parameter histories: 问答记录记录:
    ```
        histories = [
            ("question1", "answer1"),
            ("question2", "answer2"),
        ]
    ```
    """
    histories = histories or []
    history_messages = []
    for history in histories:
        history_messages.append({"role": "user", "content": history[0]})
        history_messages.append({"role": "assistant", "content": history[1]})

    if isinstance(prompt, str):
        result = [
            {"role": "system", "content": system_prompt},
        ]
        result += history_messages
        result += [
            {"role": "user", "content": prompt},
        ]
    else:
        result = prompt[:1] + history_messages + prompt[1:]
    return result


def _fix_json_escape(match):
    return "\\" + match.group(0)


def parse_json_response(response_text, default=None):
    """把LLM输出解析为json数据。

    如果LLM响应的不是有效的json数据，则抛出`ParseJsonResponseError`异常。
    """
    # 标准json输出，并且没有其它多余输出
    if response_text.startswith("```json"):
        response_text = response_text[7:]
        if response_text.endswith("```"):
            response_text = response_text[:-3]
    # s1: 尝试json反序列化
    try:
        return yaml.safe_load(response_text)
    except:
        pass
    # s2: 有多余输出，但有多个```json\nxxx\n```JSON块，将第一个有效的json块当成返回值
    json_blocks = re.findall("```json(.+?)```", response_text, re.MULTILINE)
    for json_block in json_blocks:
        try:
            return yaml.safe_load(json_block)
        except:
            pass
    # s3: 有多余输出，但有多个```\nxxx\n```代码块，将第一个有效的json块当成返回值
    json_blocks = re.findall("```json(.+?)```", response_text, re.MULTILINE)
    for json_block in json_blocks:
        try:
            return yaml.safe_load(json_block)
        except:
            pass
    # 复杂场景下，字符串没有转义，尝试修正
    response_text = re.sub('\\\\[^\\\\nrt\\"]', _fix_json_escape, response_text)
    # s4: 尝试json反序列化
    try:
        return yaml.safe_load(response_text)
    except:
        pass
    # s5: 有多余输出，但有多个```json\nxxx\n```JSON块，将第一个有效的json块当成返回值
    json_blocks = re.findall("```json(.+?)```", response_text, re.MULTILINE)
    for json_block in json_blocks:
        try:
            return yaml.safe_load(json_block)
        except:
            pass
    # s6: 有多余输出，但有多个```\nxxx\n```代码块，将第一个有效的json块当成返回值
    json_blocks = re.findall("```json(.+?)```", response_text, re.MULTILINE)
    for json_block in json_blocks:
        try:
            return yaml.safe_load(json_block)
        except:
            pass
    # 所有尝试都失败
    _logger.warning(
        "parse_json_response failed: response_text=%s",
        response_text,
    )
    return default


def chat(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    template: Optional[str] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    template_engine=None,
    template_root: Optional[str] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    **context,
):
    """基于提示词模板的对话。"""
    if (prompt is None) and (template is None):
        raise PromptOrTemplateRequired()
    original_prompt = prompt
    llm = llm or get_default_llm()
    model = model or get_default_chat_model()
    if template:
        prompt = get_template_prompt(
            template=template,
            prompt=prompt,
            template_engine=template_engine,
            template_root=template_root,
            **context,
        )
    try:
        messages = get_messages(
            prompt=prompt,
            histories=histories,
            system_prompt=system_prompt,
        )
        result = llm.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=False,
        )
        try:
            logdata = {
                "model": model,
                "messages": messages,
                "temperature": temperature,
                "max_tokens": max_tokens,
                "stream": False,
                "output": result.model_dump(),
            }
            getattr(_llm_chat_logger, _llm_chat_log_level)(
                json.dumps(logdata, ensure_ascii=False)
            )
        except Exception as error:
            _logger.exception(
                "_llm_input_output_logger failed: error=%s",
                error,
            )
        return result.choices[0].message.content
    except Exception as error:
        _logger.error(
            "chat failed: model=%s, llm=%s, original_prompt=%s, prompt=%s, histories=%s, temperature=%s, max_tokens=%s, error=%s",
            model,
            llm,
            original_prompt,
            prompt,
            histories,
            temperature,
            max_tokens,
            error,
        )
        raise ChatError()


def jsonchat(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    template: Optional[str] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    template_engine=None,
    template_root: Optional[str] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    **context,
):
    """基于提示词模板的对话。"""
    response_text = chat(
        prompt=prompt,
        histories=histories,
        template=template,
        model=model,
        llm=llm,
        template_engine=template_engine,
        system_prompt=system_prompt,
        template_root=template_root,
        temperature=temperature,
        max_tokens=max_tokens,
        **context,
    )
    return parse_json_response(
        response_text=response_text,
    )


def streaming_chat(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    template: Optional[str] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    template_engine=None,
    template_root: Optional[str] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    **context,
):
    """流式AI对话，返回响应文字内容。

    过滤掉空白块。
    """
    if (prompt is None) and (template is None):
        raise PromptOrTemplateRequired()
    original_prompt = prompt
    llm = llm or get_default_llm()
    model = model or get_default_chat_model()
    if template and isinstance(prompt, str):
        prompt = get_template_prompt(
            template=template,
            prompt=prompt,
            template_engine=template_engine,
            template_root=template_root,
            **context,
        )
    try:
        messages = get_messages(
            prompt=prompt,
            histories=histories,
            system_prompt=system_prompt,
        )
        outputs = []
        result = llm.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=True,
        )
        for chunk in result:
            empty_flag = False
            try:
                if not chunk.choices[0].delta.content:
                    empty_flag = True
            except:
                pass
            if not empty_flag:
                output = chunk.model_dump()
                yield output
                outputs.append(output)
        try:
            logdata = {
                "model": model,
                "messages": messages,
                "temperature": temperature,
                "max_tokens": max_tokens,
                "stream": True,
                "outputs": outputs,
            }
            getattr(_llm_chat_logger, _llm_chat_log_level)(
                json.dumps(logdata, ensure_ascii=False)
            )
        except Exception as error:
            _logger.exception(
                "_llm_input_output_logger failed: error=%s",
                error,
            )
    except Exception as error:
        _logger.error(
            "streaming_chat failed: model=%s, llm=%s, original_prompt=%s, prompt=%s, histories=%s, temperature=%s, max_tokens=%s, error=%s",
            model,
            llm,
            original_prompt,
            prompt,
            histories,
            temperature,
            max_tokens,
            error,
        )
        raise ChatError()


def chat_with_extra_data(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    extra_data: Optional[List[str]] = None,
    **context,
):
    return chat(
        prompt=prompt,
        histories=histories,
        model=model,
        llm=llm,
        system_prompt=system_prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        template="chat_with_extra_data.md",
        template_engine=get_template_prompt_by_jinjia2,
        template_root=os.path.join(os.path.dirname(__file__), "templates"),
        extra_data=extra_data,
        **context,
    )


def jsonchat_with_extra_data(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    extra_data: Optional[List[str]] = None,
    **context,
):
    return jsonchat(
        prompt=prompt,
        histories=histories,
        model=model,
        llm=llm,
        system_prompt=system_prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        template="chat_with_extra_data.md",
        template_engine=get_template_prompt_by_jinjia2,
        template_root=os.path.join(os.path.dirname(__file__), "templates"),
        extra_data=extra_data,
        **context,
    )


def streaming_chat_with_extra_data(
    prompt: str = None,
    histories: Optional[List[Tuple[str, str]]] = None,
    template: Optional[str] = None,
    model: Optional[str] = None,
    llm: Optional[OpenAI] = None,
    template_engine=None,
    template_root: Optional[str] = None,
    system_prompt="You are helpful assistance.",
    temperature=0.01,
    max_tokens: int = 6144,
    **context,
):
    for chunk in streaming_chat(
        prompt=prompt,
        histories=histories,
        model=model,
        llm=llm,
        system_prompt=system_prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        template="chat_with_extra_data.md",
        template_engine=get_template_prompt_by_jinjia2,
        template_root=os.path.join(os.path.dirname(__file__), "templates"),
        extra_data=extra_data,
        **context,
    ):
        yield chunk



from django.apps import AppConfig


class DjangoVectorstoreIndexModelConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "llmhelper.extra.django_vectorstore_index_model"



import time
import logging
from typing import List
from typing import Optional

from openai import OpenAI
from langchain_core.embeddings import Embeddings
from langchain_core.runnables.config import run_in_executor

from .exceptions import GetTextEmbeddingsError
from .base import get_default_llm
from .base import get_default_embeddings_model

__all__ = [
    "OpenAIEmbeddings",
    "get_text_embeddings",
]
_logger = logging.getLogger(__name__)


class OpenAIEmbeddings(Embeddings):
    """由于`langchain_openai.embeddings.OpenAIEmbeddings`无法兼容其它模型，需要重新实现适合国产embeddings模型的OpenAIEmbeddings用于向量数据库检索。"""

    def __init__(
        self,
        llm: Optional[OpenAI] = None,
        model: Optional[str] = None,
    ):
        self.llm: OpenAI = llm or get_default_llm()
        self.model: str = model or get_default_embeddings_model()

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""
        _logger.info(
            "embed_documents start: model=%s, texts=%s",
            self.model,
            texts,
        )
        stime = time.time()
        response = self.llm.embeddings.create(
            input=texts,
            model=self.model,
        )
        result = [x.embedding for x in response.data]
        used_time = time.time() - stime
        _logger.info(
            "embed_documents finished: model=%s, texts=%s, result=%s, used_time=%s",
            self.model,
            texts,
            result,
            used_time,
        )
        return result

    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""
        _logger.info(
            "embed_query start: model=%s, text=%s",
            self.model,
            text,
        )
        stime = time.time()
        response = self.llm.embeddings.create(
            input=text,
            model=self.model,
        )
        result = response.data[0].embedding
        used_time = time.time() - stime
        _logger.info(
            "embed_query finished: model=%s, text=%s, result=%s, used_time=%s",
            self.model,
            text,
            result,
            used_time,
        )
        return result

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous Embed search docs."""
        return await run_in_executor(None, self.embed_documents, texts)

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text."""
        return await run_in_executor(None, self.embed_query, text)


def get_text_embeddings(
    text,
    llm: Optional[OpenAI] = None,
    model: str = None,
):
    """将文字转化为向量用于向量数据库检索。向量以浮点数数组表示。"""
    llm = llm or get_default_llm()
    model = model or get_default_embeddings_model()
    _logger.info(
        "get_text_embeddings start: model=%s, text=%s",
        model,
        text,
    )
    stime = time.time()
    try:
        result = (
            llm.embeddings.create(
                input=text,
                model=model,
            )
            .data[0]
            .embedding
        )
    except Exception as error:
        _logger.warning(
            "get_text_embeddings failed: model=%s, text=%s, error=%s",
            model,
            text,
            error,
        )
        raise GetTextEmbeddingsError()
    used_time = time.time() - stime
    _logger.info(
        "get_text_embeddings finished: model=%s, text=%s, result=%s, used_time=%s",
        model,
        text,
        result,
        used_time,
    )
    return result



import time
import logging
from typing import Optional

import requests

from openai import OpenAI

from .exceptions import GetRerankScoresError
from .exceptions import RerankNoneValueError
from .base import get_default_llm
from .base import get_default_rerank_model
from .base import get_llm_api_url

__all__ = [
    "get_rerank_scores",
]
_logger = logging.getLogger(__name__)


def get_rerank_scores(
    query,
    documents,
    llm: Optional[OpenAI] = None,
    model: Optional[str] = None,
    max_size: int = 1024,
):
    """返回文本相似度得分列表。"""
    if not documents:
        return []
    if query is None:
        raise RerankNoneValueError()
    for doc in documents:
        if doc is None:
            raise RerankNoneValueError()
    llm = llm or get_default_llm()
    model = model or get_default_rerank_model()
    # 如果query, documents字符串超过最大长度，则截取最前面的字符串
    if len(query) > max_size:
        _logger.warning(
            "get_rerank_scores: the query string exceeds the limit, max_size=%s, query=%s",
            max_size,
            query,
        )
        query = query[:max_size]
    fixed_documents = []
    for document in documents:
        if len(document) > max_size:
            _logger.warning(
                "get_rerank_scores: the document string exceeds the limit, max_size=%s, query=%s, document=%s",
                max_size,
                query,
                document,
            )
            document = document[:max_size]
        fixed_documents.append(document)
    # 构建并请求
    stime = time.time()
    url = get_llm_api_url("rerank", llm=llm)
    headers = {
        "Authorization": "Bearer " + llm.api_key,
        "accept": "application/json",
        "Content-Type": "application/json",
    }
    data = {
        "model": model,
        "query": query,
        "documents": fixed_documents,
    }
    _logger.debug(
        "get_rerank_scores request start: url=%s, headers=%s, data=%s",
        url,
        headers,
        data,
    )
    response = requests.post(
        url,
        json=data,
        headers=headers,
    )
    _logger.debug(
        "get_rerank_scores request finished: url=%s, headers=%s, data=%s, response=%s",
        url,
        headers,
        data,
        response.text,
    )
    result = [
        x["relevance_score"]
        for x in sorted(response.json().get("results", []), key=lambda z: z["index"])
    ]
    used_time = time.time() - stime
    if not result:
        _logger.error(
            "get_rerank_scores request failed: url=%s, headers=%s, data=%s, response=%s",
            url,
            headers,
            data,
            response.text,
        )
        raise GetRerankScoresError()
    return result



import os
from typing import Optional

import yaml
from openai import OpenAI

__all__ = [
    "get_llmhelper_config",
    "set_llmhelper_default_config",
    "set_llmhelper_config",
    "get_default_llm",
    "get_default_chat_model",
    "get_default_embeddings_model",
    "get_default_rerank_model",
    "get_default_redis_stack_url",
    "get_template_engine",
    "get_llm_base_url",
    "get_llm_api_url",
    "get_llm_chat_log_name",
    "get_llm_chat_log_level",
]

OPENAI_DEFAULT_CHAT_MODEL = "qwen2-instruct"
OPENAI_DEFAULT_EMBEDDINGS_MODEL = "bge-m3"
OPENAI_DEFAULT_RERANK_MODEL = "bge-reranker-v2-m3"
LLMHELPER_DEFAULT_REDIS_STACK_URL = "redis://localhost:6379/0"
LLMHELPER_CONFIG = {}


def get_llmhelper_config():
    """获取llmehlper设置。"""
    return LLMHELPER_CONFIG


def set_llmhelper_default_config(
    api_key=None,
    base_url=None,
    chat_model=None,
    embeddings_model=None,
    rerank_model=None,
    template_engine=None,
    redis_stack_url=None,
    llm_chat_log_name=None,
    llm_chat_log_level=None,
    model=None,  # alias for chat_model
):
    """设置llmhelper全局参数默认值。

    即之前没有设定，则设置。
    如果之前已经设置，则忽略。
    """
    chat_model = chat_model or model
    if api_key:
        LLMHELPER_CONFIG.setdefault("api_key", api_key)
    if base_url:
        LLMHELPER_CONFIG.setdefault("base_url", base_url)
    if chat_model:
        LLMHELPER_CONFIG.setdefault("chat_model", chat_model)
    if embeddings_model:
        LLMHELPER_CONFIG.setdefault("embeddings_model", embeddings_model)
    if rerank_model:
        LLMHELPER_CONFIG.setdefault("rerank_model", rerank_model)
    if template_engine:
        LLMHELPER_CONFIG.setdefault("template_engine", template_engine)
    if redis_stack_url:
        LLMHELPER_CONFIG.setdefault("redis_stack_url", redis_stack_url)
    if llm_chat_log_name:
        LLMHELPER_CONFIG.setdefault("llm_chat_log_name", llm_chat_log_name)
    if llm_chat_log_level:
        LLMHELPER_CONFIG.setdefault("llm_chat_log_level", llm_chat_log_level)


def set_llmhelper_config(
    api_key=None,
    base_url=None,
    chat_model=None,
    embeddings_model=None,
    rerank_model=None,
    template_engine=None,
    redis_stack_urls=None,
    llm_chat_log_name=None,
    llm_chat_log_level=None,
    model=None,  # alias for chat_model
):
    """设置llmhelper全局参数。"""
    chat_model = chat_model or model
    if api_key:
        LLMHELPER_CONFIG["api_key"] = api_key
    if base_url:
        LLMHELPER_CONFIG["base_url"] = base_url
    if chat_model:
        LLMHELPER_CONFIG["chat_model"] = chat_model
    if embeddings_model:
        LLMHELPER_CONFIG["embeddings_model"] = embeddings_model
    if rerank_model:
        LLMHELPER_CONFIG["rerank_model"] = rerank_model
    if template_engine:
        LLMHELPER_CONFIG["template_engine"] = template_engine
    if redis_stack_urls:
        LLMHELPER_CONFIG["redis_stack_urls"] = redis_stack_urls
    if llm_chat_log_name:
        LLMHELPER_CONFIG["llm_chat_log_name"] = llm_chat_log_name
    if llm_chat_log_level:
        LLMHELPER_CONFIG["llm_chat_log_level"] = llm_chat_log_level


def get_default_llm():
    """创建系统默认的OpenAI实例。"""
    api_key = LLMHELPER_CONFIG.get("api_key", os.environ.get("OPENAI_API_KEY"))
    base_url = LLMHELPER_CONFIG.get("base_url", os.environ.get("OPENAI_BASE_URL"))
    return OpenAI(
        api_key=api_key,
        base_url=base_url,
    )


def get_default_chat_model():
    """获取系统默认的对话模型名称。"""
    return LLMHELPER_CONFIG.get(
        "chat_model",
        os.environ.get(
            "OPENAI_CHAT_MODEL",
            os.environ.get(
                "OPENAI_MODEL",
                OPENAI_DEFAULT_CHAT_MODEL,
            ),
        ),
    )


def get_default_embeddings_model():
    """获取系统默认的嵌入模型名称。"""
    return LLMHELPER_CONFIG.get(
        "embeddings_model",
        os.environ.get(
            "OPENAI_EMBEDDINGS_MODEL",
            OPENAI_DEFAULT_EMBEDDINGS_MODEL,
        ),
    )


def get_default_rerank_model():
    """获取系统默认的嵌入模型名称。"""
    return LLMHELPER_CONFIG.get(
        "rerank_model",
        os.environ.get(
            "OPENAI_RERANK_MODEL",
            OPENAI_DEFAULT_RERANK_MODEL,
        ),
    )


def get_default_redis_stack_url(index_name=None):
    """获取系统默认的向量数据库地址。

    默认使用redis-stack作为向量数据库。
    """
    index_name = index_name or "default"
    config = LLMHELPER_CONFIG.get(
        "redis_stack_urls",
        yaml.safe_load(
            os.environ.get(
                "LLMHELPER_REDIS_STACK_URLS",
                "{}",
            )
        ),
    )
    index_parts = index_name.split(":")
    for index in range(len(index_parts)):
        config_key = ":".join(index_parts[: len(index_parts) - index])
        if config_key in config:
            return config[config_key]
    return config.get("default", LLMHELPER_DEFAULT_REDIS_STACK_URL)


def get_template_engine():
    """获取系统提示词模板引擎。"""
    return LLMHELPER_CONFIG.get("template_engine", None)


def get_llm_base_url(
    llm: Optional[OpenAI] = None,
):
    """返回规范化的OPEN AI BASE_URL。

    规范化的URL格式为：http://host/v1/。
    """
    llm = llm or get_default_llm()
    url = str(llm.base_url)
    if not url.endswith("/"):
        url += "/"
    if not url.endswith("v1/"):
        url += "v1/"
    return url


def get_llm_api_url(
    name,
    llm: Optional[OpenAI] = None,
):
    """获取OPENAI接口指定服务的URL地址。"""
    url = get_llm_base_url(llm=llm)
    return url + name


def get_llm_chat_log_name():
    return LLMHELPER_CONFIG.get(
        "get_llm_chat_log_name",
        os.environ.get(
            "LLM_CHAT_LOG_NAME",
            "llmhelper.llm",
        ),
    )


def get_llm_chat_log_level():
    return LLMHELPER_CONFIG.get(
        "llm_chat_log_level",
        os.environ.get(
            "LLM_CHAT_LOG_LEVEL",
            "warning",
        ),
    )



from typing import Optional
import logging

from zenutils import strutils
import pydantic
import redis
from redis import Redis
from langchain_community.vectorstores.redis import Redis as LangchainRedisVectorStore
from openai import OpenAI

from .exceptions import VectorStoreNoneValueError
from .base import get_default_llm
from .base import get_default_embeddings_model
from .base import get_default_rerank_model
from .base import get_default_redis_stack_url
from .embeddings import OpenAIEmbeddings
from .rerank import get_rerank_scores

__all__ = [
    "RedisVectorStoreMetaBase",
    "RedisVectorStore",
    "get_cached_vectorstore",
]
_logger = logging.getLogger(__name__)


class RedisVectorStoreMetaBase(pydantic.BaseModel):
    """Redis向量数据库meta基础类"""

    vs_uid: Optional[str] = None
    vs_page_content: Optional[str] = None
    vs_embeddings_score: Optional[float] = None
    vs_rerank_score: Optional[float] = None


class RedisVectorStore(object):
    """基于redis-stack的向量数据库。"""

    def __init__(
        self,
        index_name=None,
        redis_stack_url: str = None,
        llm: Optional[OpenAI] = None,
        model: Optional[str] = None,
        embeddings_model: Optional[str] = None,
        rerank_model: Optional[str] = None,
        max_size: Optional[int] = 1024,
    ):
        # 获取参数或获取默认参数
        self.max_size: int = max_size or 1024
        self.index_name: str = index_name
        self.redis_stack_url = redis_stack_url or get_default_redis_stack_url(
            index_name=index_name
        )
        self.embeddings_model = (
            embeddings_model or model or get_default_embeddings_model()
        )
        self.rerank_model = rerank_model or get_default_rerank_model()
        # 构建AI实例
        self.llm = llm or get_default_llm()
        self.store: LangchainRedisVectorStore = None
        self.redis: Redis = None
        # 加载实例
        self.redis = redis.from_url(self.redis_stack_url)
        self.embeddings = OpenAIEmbeddings(
            llm=self.llm,
            model=self.embeddings_model,
        )
        if index_name:
            self.store = LangchainRedisVectorStore(
                redis_url=self.redis_stack_url,
                index_name=self.index_name,
                key_prefix=self.index_name,
                embedding=self.embeddings,
            )

    def safe_insert(self, content, meta=None):
        if content is None:
            raise VectorStoreNoneValueError()
        if not self.store:
            raise RuntimeError("未指定index_name的VectorStore实例不允许新建索引")
        contents = strutils.chunk(content, self.max_size)
        if meta:
            return self.store.add_texts(contents, [meta] * len(contents))
        else:
            return self.store.add_texts(contents)

    def insert(self, content, meta=None):
        if content is None:
            raise VectorStoreNoneValueError()
        if not self.store:
            raise RuntimeError("未指定index_name的VectorStore实例不允许新建索引")
        if len(content) > self.max_size:
            _logger.warning(
                "Insert content length exceeds the limit, truncate to the limit length. size=%s, max_size=%s",
                len(content),
                self.max_size,
            )
            content = content[: self.max_size]
        if meta:
            return self.store.add_texts([content], [meta])
        else:
            return self.store.add_texts([content])

    def insert_many(self, contents, metas=None):
        for content in contents:
            if content is None:
                raise VectorStoreNoneValueError()
        if not self.store:
            raise RuntimeError("未指定index_name的VectorStore实例不允许新建索引")
        safe_contents = []
        for content in contents:
            if len(content) > self.max_size:
                _logger.warning(
                    "Insert content length exceeds the limit, truncate to the limit length. size=%s, max_size=%s",
                    len(content),
                    self.max_size,
                )
                content = content[: self.max_size]
            safe_contents.append(content)
        if metas:
            return self.store.add_texts(contents, metas)
        else:
            return self.store.add_texts(contents)

    def delete(self, uid):
        return self.redis.delete(uid)

    def delete_many(self, uids):
        if uids:
            return self.redis.delete(*uids)
        else:
            return 0

    def similarity_search_with_relevance_scores(
        self,
        query,
        k=4,
        raise_on_error=False,
        **kwargs,
    ):
        if query is None:
            _logger.exception(
                "similarity_search_with_relevance_scores got None query..."
            )
            return []
        try:
            docs = []
            for (
                doc,
                vs_embeddings_score,
            ) in self.store.similarity_search_with_relevance_scores(
                query=query, k=k, **kwargs
            ):
                docs.append(
                    {
                        "vs_uid": doc.metadata["id"],
                        "vs_page_content": doc.page_content,
                        "vs_embeddings_score": vs_embeddings_score,
                    }
                )
            for doc in docs:
                info = self.redis.hgetall(doc["vs_uid"])
                for k, v in info.items():
                    if isinstance(k, bytes):
                        k = k.decode("utf-8")
                    if k in ["content_vector"]:
                        continue
                    if isinstance(v, bytes):
                        v = v.decode("utf-8")
                    if not v:
                        v = None
                    doc[k] = v
            docs.sort(
                key=lambda x: -x["vs_embeddings_score"],
            )
            return docs
        except Exception as error:
            _logger.exception(
                "similarity_search_with_relevance_scores failed: error=%s, query=%s",
                error,
                query,
            )
            if raise_on_error:
                raise error
            else:
                return []

    def search_and_rerank(
        self,
        query,
        k=4,
        scale=4,
        reranker_score_threshold=None,
        raise_on_error=False,
        **kwargs,
    ):
        """搜索并重排

        @parameter: scale，表示一次搜索的放大位数，即一次搜索时获取的文档数量为[k*scale]个。
        """
        try:
            scale = int(scale)
            if scale < 1:
                scale = 1
            docs = self.similarity_search_with_relevance_scores(
                query=query,
                k=k * scale,
                **kwargs,
            )
            documents = [doc["vs_page_content"] for doc in docs]
            scores = get_rerank_scores(
                query=query,
                documents=documents,
                llm=self.llm,
                model=self.rerank_model,
            )
            for doc, score in zip(docs, scores):
                doc["vs_rerank_score"] = score
            docs.sort(
                key=lambda x: -x["vs_rerank_score"],
            )
            if reranker_score_threshold:
                docs = [
                    doc
                    for doc in docs
                    if doc["vs_rerank_score"] >= reranker_score_threshold
                ]
            return docs[:k]
        except Exception as error:
            _logger.exception(
                "search_and_rerank failed: error=%s, query=%s",
                error,
                query,
            )
            if raise_on_error:
                raise error
            else:
                return []

    def flush(self):
        """清空指定索引。"""
        # 删除所有索引项
        keys = self.redis.keys(self.index_name + ":*")
        if keys:
            self.redis.delete(*keys)
        # 删除索引
        indexes = self.redis.execute_command("FT._LIST")
        if indexes:
            indexes = [x.decode("utf-8") for x in indexes]
        if self.index_name in indexes:
            self.redis.execute_command(f"FT.DROPINDEX {self.index_name}")
        return len(keys)


GLOBAL_CACHED_VECTORSTORES = {}


def get_cached_vectorstore(
    index_name=None,
    redis_stack_url: str = None,
    llm: Optional[OpenAI] = None,
    model: Optional[str] = None,
    embeddings_model: Optional[str] = None,
    rerank_model: Optional[str] = None,
) -> RedisVectorStore:
    """获取缓存过的Redis向量数据库实例"""
    cache_key = f"{index_name}:{model}:{embeddings_model}:{rerank_model}"

    if index_name in GLOBAL_CACHED_VECTORSTORES:
        return GLOBAL_CACHED_VECTORSTORES[cache_key]

    rvs = RedisVectorStore(
        index_name=index_name,
        redis_stack_url=redis_stack_url,
        llm=llm,
        model=model,
        embeddings_model=embeddings_model,
        rerank_model=rerank_model,
    )
    GLOBAL_CACHED_VECTORSTORES[cache_key] = rvs
    return GLOBAL_CACHED_VECTORSTORES[cache_key]
